import re
import numpy as np
import datetime
from collections import Counter

from pyg_base import is_bool, try_back, replace, alphabet, ALPHABET, is_str, \
    as_list, logger, tree_repr, NoneType
from pyg_mongo._encoders import encode
from operator import and_, or_
from functools import reduce

__all__ = ['Q','q', 'mdict']


_id = '_id'
_js = 'js'
_regex = 'regex'
_bin = 'bin'
_and = '$and'
_or = '$or'
_not_in = '$nin'
_eq = '$eq'
_ne = '$ne'
_ge = '$gte'
_le = '$lte'
_gt = '$gt'
_lt = '$lt'
_not = '$not'
_mod = '$mod'
_set = '$set'
_unset = '$unset'
_rename = '$rename'
_exists = '$exists'
_in = '$in'
_not_in = '$nin'
_type = '$type'
_prev = '_prev'
_data = 'data'
_doc = 'doc'
_deleted = 'deleted'



_type2bson = {float : [1,19], 
              str : 2,
              np.nan : 10, 
              np.ndarray : 4, 
              _id : 7, 
              _js : [13,15], 
              bool : 8, 
              None: 10, 
              _regex : 11,
              _bin : 5,
              datetime.datetime: [9,17],
              datetime.date: 9,
              int : [16,18]
              }

_sorted = try_back(sorted)

def _is_query(query):
    if isinstance(query, mdict):
        return True
    for key, value in query.items():
        if key.startswith('$'):
            return True
        elif isinstance(value, dict):
            if _is_query(value):
                return True
    return False


def _as_key(key):
    txt = ''.join([char if char in alphabet + ALPHABET else '_' for char in key])
    txt = replace(txt, '__','_')
    if txt.startswith('_') and not key.startswith('_'):
        txt = txt[1:]
    if txt.endswith('_') and not key.endswith('_'):
        txt = txt[:-1]
    return txt

def _case_proxies(keys):
    """
    creates a dict of possible keys mapping to original keys
    
    The reason is to allow getattr on keys with invalid characters in them
    
    :Example:
    ---------
    >>> from pyg import Q
    >>> query = Q(['hello world', 'James Joyce'])
    >>> (query.hello_world == 1) & (query.james_joyce == 2)
    >>> {"$and": [{"James Joyce": {"$eq": 2}}, {"hello world": {"$eq": 1}}]}
    
    
    :Parameters:
    ----------------
    keys : list of strings
        keys for which we will use a proxy to specify these keys


    """
    keys = set(as_list(keys))
    res = {key: key for key in keys}
    proxy = [_as_key(str(k)) for k in keys]
    counter = Counter(proxy)    
    res.update({p : value for p, value in zip(proxy, keys) if counter[p] == 1})
    proxy = [p.lower() for p in proxy]
    counter = Counter(proxy)    
    res.update({p : value for p, value in zip(proxy, keys) if counter[p] == 1})
    proxy = [p.upper() for p in proxy]
    counter = Counter(proxy)    
    res.update({p : value for p, value in zip(proxy, keys) if counter[p] == 1})
    return res
    

def _as_items(query):
    """
    converts a query into keys and tuples
    """
    if query is None:
        return []
    elif isinstance(query, dict):
        return _sorted([(key, _as_items(value)) for key, value in query.items()])
    elif isinstance(query, list):
        return _sorted([_as_items(key) for key in query])
    else:
        return query


def _q_set(values):
    if len(values) > 1:
        values = _sorted(values, key = _as_items)
        items = [_as_items(v) for v in values]
        values = [values[i] for i in range(len(values)) if i ==0 or items[i]!=items[i-1]]
        return values[1:] if items[0] == [] else values
    elif len(values) == 1:
        return [] if _as_items(values[0]) == [] else values
    else:
        return values

def _q_and(values):
    values = _q_set(values)
    return reduce(and_, values) if len(values)>0 else mdict()

def _q_or(values):
    values = _q_set(values)
    return reduce(or_, values) if len(values)>0 else mdict()


class mkey(object):
    """
    mongo query key
    """
    def __init__(self, key):
        self._key = key

    def __getattr__(self, subkey):
        return mkey('%s.%s'%(self._key, subkey))
    
    def __getitem__(self, subkey):
        if is_str(subkey):
            return mkey('%s.%s'%(self._key, subkey))
        elif isinstance(subkey, slice):
            if subkey.start is None and subkey.stop is None:
                return self
            elif subkey.start is None:
                return self < subkey.stop
            elif subkey.stop is None:
                return self >= subkey.start
            else:
                return (self >= subkey.start) & (self < subkey.stop)
        else:
            return self == subkey

    def __call__(self, **kwargs):
        return _q_and([self[key] == value for key, value in kwargs.items()])
    
    def _set(self, other):
        other = encode(other)
        return mdict({self._key : other})
    
    def __pos__(self):
        return self._set({_exists : True})

    def __neg__(self):
        return self._set({_exists : False})

    __invert__ = __neg__
    
    @property 
    def exists(self):
        return self._set({_exists : True})

    @property 
    def not_exists(self):
        return self._set({_exists : False})

    def __eq__(self, other):
        if isinstance(other, re.Pattern):
            return self._set({_regex : other.pattern})
        other = encode(other)
        if isinstance(other, list):
            if len(other) == 1:
                return self._set({_eq: other[0]})
            else:
                return self._set({_in : other})
        elif is_bool(other):
            return self._set({_in : [True, 1] if other else [False, 0]})
        elif isinstance(other, dict):
            return self(**other)
        else:
            return self._set({_eq: other}) 
    
    def __ne__(self, other):
        if isinstance(other, re.Pattern):
            return self._set({_not : {_regex : other.pattern}})
        elif isinstance(other, list):
            if len(other) == 1:
                return self._set({_ne : other})
            else:
                return self._set({_not_in : other})
        else:
            return self._set({_ne : other})
    
    def __ge__(self, other):
        return self._set({ _ge : other })

    def __gt__(self, other):
        return self._set({ _gt : other })

    def __le__(self, other):
        return self._set({ _le : other })
    
    def __lt__(self, other):
        return self._set({ _lt : other })

    def __mod__(self, other):
        return modkey(self._key, other)

    def __and__(self, other) : 
        return (+self) & other

    def __add__(self, other) : 
        return (+self) + other

    def __sub__(self, other) : 
        return (+self) & (~other)

    def __or__(self, other) : 
        return (+self) | other
            
    def isinstance(self, *types):
        """
         see https://docs.mongodb.com/manual/reference/operator/type  
        """
        bson_types = _sorted(set(sum([as_list(_type2bson.get(t)) for t in types],[])))
        return self._set({_type : bson_types})


class modkey(object):

    def __init__(self, key, div):
        self._key = key
        self._div = div

    def __eq__(self, remainder):
        return mdict({self._key: {_mod : [self._div, remainder]}})

    def __ne__(self, remainder):
        return mdict({self._key: { _not: {_mod : [self._div, remainder]}}})
    
    def __pos__(self):
        return self.__ne__(0)
    
    def __neg__(self):
        return self.__eq__(0)
    
    __invert__ = __neg__
    
    def __and__(self, other) : 
        return (+self) & other

    def __add__(self, other) : 
        return (+self) + other

    def __sub__(self, other) : 
        return (+self) - other

    def __or__(self, other) : 
        return (+self) | other

    def __repr__(self) : 
        return 'mod(%s, %s)'%(self._key, self._div)
    
class mdict(dict):
    def _chain(self, other, key):
        rollback = False
        if not isinstance(other, dict):
            _mkey = other
            other = +other
            rollback = True
        if list(self.keys()) == [key]:
            res = mdict({key : list(self[key])})
            res[key].append(other)
        elif list(other.keys()) == [key]:
            res = mdict({key : list(other[key])})
            res[key].append(self)
        else:
            res = mdict({key : [self, other]})
        if rollback:
            res._mkey = _mkey
            res._prev = self
            res._key = key
        res[key] = _q_set(res[key])
        if len(res[key]) == 1:
            return self
        else:
            return res
    
    def __and__(self, other):
        return self._chain(other, _and)
    
    __add__ = __and__ 
    
    update = __and__
    
    def __sub__(self, other):
        return self + (-other)
    
    def __or__(self, other):
        return self._chain(other, _or)
    
    def __pos__(self): 
        return self

    def __neg__(self):
        return mdict({_not: self})
    
    __invert__ = __neg__
    
    def __eq__(self, other):
        if hasattr(self, _prev):
            logger.warning('WARN: operation order ambiguity, you should use brackets')
            return self._prev._chain(self._mkey == other, self._key)
        else:
            return super(mdict, self).__eq__(other)

    def __ne__(self, other):
        if hasattr(self, _prev):
            logger.warning('WARN: operation order ambiguity, you should use brackets')
            return self._prev._chain(self._mkey != other, self._key)
        else:
            return super(mdict, self).__ne__(other)

    def __ge__(self, other):
        if hasattr(self, _prev):
            logger.warning('WARN: operation order ambiguity, you should use brackets')
            return self._prev._chain(self._mkey>=other, self._key)
        else:
            raise ValueError('cannot compare dicts this way')

    def __gt__(self, other):
        if hasattr(self, _prev):
            logger.warning('WARN: operation order ambiguity, you should use brackets')
            return self._prev._chain(self._mkey > other, self._key)
        else:
            raise ValueError('cannot compare dicts this way')


    def __le__(self, other):
        if hasattr(self, _prev):
            logger.warning('WARN: operation order ambiguity, you should use brackets')
            return self._prev._chain(self._mkey<=other, self._key)
        else:
            raise ValueError('cannot compare dicts this way')

    def __lt__(self, other):
        if hasattr(self, _prev):
            logger.warning('WARN: operation order ambiguity, you should use brackets')
            return self._prev._chain(self._mkey < other, self._key)
        else:
            raise ValueError('cannot compare dicts this way')

    def __repr__(self):
        res = tree_repr(dict(self))
        return res.replace('True', 'true').replace('False', 'false').replace("'", '"')

    def __str__(self):
        return 'M%s'%(dict(self).__str__().replace('M{', '{'))
    

class Q(dict):
    """
    
    The MongoDB interface for query of a collection (table) is via a creation of a complicated looking dict:
    https://docs.mongodb.com/manual/tutorial/query-documents/
    
    This is rather complicated for the average user so Q simplifies it greatly.
    Q is based on TinyDB and users of TinyDB will recognise it. 
    https://tinydb.readthedocs.io/en/latest/usage.html
    
    q is the singleton of Q.
    
    q supports both *calling* to generate the querying dict

    >>> q(a = 1, b = 2)

    or

    >>> (q.a == 1) & (q.b == 2)  # {"$and": [{"a": {"$eq": 1}}, {"b": {"$eq": 2}}]}
    >>> (q.a == 1) | (q.b == 2)  # {"$or": [{"a": {"$eq": 1}}, {"b": {"$eq": 2}}]}

    or indeed 
    
    >>> q(q.a == 1, q.b  == 2)
       
    :Example:
    -------
    >>> from pyg import q
    >>> import re
    
    >>> assert dict(q.a == 1) == {"a": {"$eq": 1}}
    >>> assert dict(q(a = [1,2])) == {'a': {'$in': [1, 2]}}
    >>> assert dict(q(q.a == [1,2], q.b > 3)) == {'$and': [{"a": {"$in": [1, 2]}}, {"b": {"$gt": 3}}]}  # a in [1,2] and b greater than 3
    >>> assert dict(q(a = re.compile('^hello'))) == {'a': {'regex': '^hello'}}     # a regex query using regex expressions    
    
    Existence
    ---------------
    >>> assert dict(q.a.exists + q.b.not_exists)  == {"$and": [{"a": {"$exists": true}}, {"b": {"$exists": false}}]}

    Not operator
    ------------------
    >>> assert dict(~(q.a==1))  == {'$not': {"a": {"$eq": 1}}}


    """
        
    def __init__(self, keys = None):
        if hasattr(keys, 'keys'):
            keys = getattr(keys, 'keys')
        if hasattr(keys, 'columns'):
            keys = getattr(keys, 'columns')
        if callable(keys):
            keys = keys()
        super(Q, self).__init__(_case_proxies(as_list(keys)))
    
    def __dir__(self):
        return _sorted(self.keys())
    
    def __getattr__(self, value):
        return mkey(super(Q, self).__getitem__(value)) if len(self) else mkey(value)
    
    def __call__(self, *args, **kwargs):
        values = [self[key] == value for key, value in kwargs.items()] + [self[arg] for arg in args]
        return _q_and(values)
    
    def __getitem__(self, value):
        if isinstance(value, list) and min([isinstance(v, (dict, list, NoneType)) for v in value], default = True):
            values = [self[v] for v in value]
            return _q_or(values)
        elif isinstance(value, dict):
            if _is_query(value):
                return value
            else:
                return self(**value)
        elif value is None:
            return mdict()
        else:
            return mkey(super(Q, self).__getitem__(value)) if len(self) else mkey(value)

q = Q()
                        


from pyg_base import is_str, cfg_read, get_cache
from pyg_mongo._reader import mongo_reader
from pyg_mongo._cursor import mongo_cursor, mongo_pk_cursor


from pymongo import MongoClient

__all__ = ['mongo_table']


def _url(url):
    """
    converts the URL address to actual url based on the cfg['mongo'] locations. see cfg_read() for help.
    """
    cfg = cfg_read()
    mongo = cfg.get('mongo', {})
    _url = url or 'null'
    if _url in mongo:
        return mongo[_url]
    else:
        return url
    

_mongo_table_cache = get_cache('mongo_table') 
_mongo_table_cache['w'] = mongo_cursor, mongo_pk_cursor, MongoClient
_mongo_table_cache['r'] = mongo_reader, mongo_reader, MongoClient


def mongo_table(table, db, pk = None, url = None, reader = None, writer = None, mode = 'w', **kwargs):    
    """
    :Example:
    ---------
    from pyg import *
    from pymongo import MongoClient
    from motor import MotorClient

    table = db = 'test'
    pk = 'key'
    url = reader = writer = None    
    mode = 'aw'    
    kwargs = {}
    client = MotorClient(url)
    c = client[db][table]

    isinstance(cursor, ())
    res = obj(c, pk = pk, writer = writer, reader = reader, **kwargs)
    
    await c.insert_one(dict(a = 1))
    await c.create_index(dict(a = 1))

    cfg = cfg_read()
    
    """ 
    if mode is None:
        mode = 'w'
    if is_str(mode):
        mode = mode.lower()
        if_no_pk, if_pk, client = _mongo_table_cache[mode]
        obj = if_no_pk if pk is None else if_pk
    else:
        obj = mode
        client = MongoClient
    url = _url(url)
    c = client(url)[db][table]
    res = obj(c, pk = pk, writer = writer, reader = reader, **kwargs)
    if isinstance(res, (mongo_reader)) and len(res) == 0:
         res.create_index()
    return res

        



import pandas as pd
import numpy as np
from bson.objectid import ObjectId
from pyg_base import pd_to_parquet, pd_read_parquet, pd_to_npy, np_save, pd_read_npy, is_pd, is_dict, is_series, is_arr, is_date, dt2str, mkdir, tree_items
from pyg_base import encode as encode_
from functools import partial


_parquet = '.parquet'
_npy = '.npy'
_csv = '.csv'
_series = '_is_series'
_root = 'root'
_db = 'db'
_obj = '_obj'

def encode(value):
    """
    pyg_mongo.encoder is similar to pyg_base.encoder with the only exception being that bson.objectid.ObjectId used by mongodb to generate the document _id, are not encoded

    Parameters
    ----------
    value : value/document to be encoded

    Returns
    -------
    encoded value/document
    """
    return encode_(value, ObjectId)


__all__ = ['root_path', 'pd_to_csv', 'pd_read_csv', 'parquet_encode', 'parquet_write', 'csv_encode', 'csv_write']

def root_path(doc, root, fmt = None):
    """
    returns a location based on doc
    
    :Example:
    --------------
    >>> root = 'c:/%school/%pupil.name/%pupil.surname/'
    >>> doc = dict(school = 'kings', 
                   pupil = dict(name = 'yoav', surname = 'git'), 
                   grades = dict(maths = 100, physics = 20, chemistry = 80), 
                   report = dict(date = dt(2000,1,1), 
                                 teacher = dict(name = 'adam', surname = 'cohen')
                                 )
                   )
    
    >>> assert root_path(doc, root) == 'c:/kings/yoav/git/'

    The scheme is entirely up to the user and the user needs to ensure what defines the unique primary keys that prevent documents overstepping each other...
    >>> root = 'c:/%school/%pupil.name_%pupil.surname/'
    >>> assert root_path(doc, root) == 'c:/kings/yoav_git/'
    
    >>> root = 'c:/archive/%report.date/%pupil.name.%pupil.surname/'
    >>> assert root_path(doc, root, '%Y') == 'c:/archive/2000/yoav.git/'  # can choose to format dates by providing a fmt.
    """
    items = sorted(tree_items(dict(doc)))[::-1]
    res = root
    for row in items:
        text = '%(' + '.'.join(row[:-1]) + ')'
        if text in root:
            value = dt2str(row[-1], fmt).replace(':','') if is_date(row[-1]) else str(row[-1])
            res = res.replace(text, '%s'% value)
        text = '%' + '.'.join(row[:-1])
        if text in root:
            value = dt2str(row[-1], fmt).replace(':','') if is_date(row[-1]) else str(row[-1])
            res = res.replace(text, '%s'% value)
    if '%' in res:
        raise ValueError('The document did not contain enough keys to determine the path %s'%res)
    return res

def pd_to_csv(value, path):
    """
    A small utility to write both pd.Series and pd.DataFrame to csv files
    """
    assert is_pd(value), 'cannot save non-pd'
    if is_series(value):
        value.index.name = _series
    if value.index.name is None:
        value.index.name = 'index'
    if path[-4:].lower()!=_csv:
        path = path + _csv
    mkdir(path)
    value.to_csv(path)
    return path

def pd_read_csv(path):
    """
    A small utility to read both pd.Series and pd.DataFrame from csv files
    """
    res = pd.read_csv(path)
    if res.columns[0] == _series and res.shape[1] == 2:
        res = pd.Series(res[res.columns[1]], res[_series].values)
        return res
    if res.columns[0] == 'index':
        res = res.set_index('index')
    return res

_pd_read_csv = encode(pd_read_csv)
_pd_read_parquet = encode(pd_read_parquet)
_pd_read_npy = encode(pd_read_npy)
_np_load = encode(np.load)

def parquet_encode(value, path, compression = 'GZIP'):
    """
    encodes a single DataFrame or a document containing dataframes into a an abject that can be decoded

    >>> from pyg import *     
    >>> path = 'c:/temp'
    >>> value = dict(key = 'a', n = np.random.normal(0,1, 10), data = dictable(a = [pd.Series([1,2,3]), pd.Series([4,5,6])], b = [1,2]), other = dict(df = pd.DataFrame(dict(a=[1,2,3], b= [4,5,6]))))
    >>> encoded = parquet_encode(value, path)
    >>> assert encoded['n']['file'] == 'c:/temp/n.npy'
    >>> assert encoded['data'].a[0]['path'] == 'c:/temp/data/a/0.parquet'
    >>> assert encoded['other']['df']['path'] == 'c:/temp/other/df.parquet'

    >>> decoded = decode(encoded)
    >>> assert eq(decoded, value)

    """
    if path.endswith(_parquet):
        path = path[:-len(_parquet)]
    if path.endswith('/'):
        path = path[:-1]
    if is_pd(value):
        return dict(_obj = _pd_read_parquet, path = pd_to_parquet(value, path + _parquet))
    elif is_arr(value):
        mkdir(path + _npy)
        np.save(path + _npy, value)
        return dict(_obj = _np_load, file = path + _npy)        
    elif is_dict(value):
        return type(value)(**{k : parquet_encode(v, '%s/%s'%(path,k), compression) for k, v in value.items()})
    elif isinstance(value, (list, tuple)):
        return type(value)([parquet_encode(v, '%s/%i'%(path,i), compression) for i, v in enumerate(value)])
    else:
        return value
    
def npy_encode(value, path, append = False):
    """
    >>> from pyg_base import * 
    >>> value = pd.Series([1,2,3,4], drange(-3))

    """
    if path.endswith(_npy):
        path = path[:-len(_npy)]
    if path.endswith('/'):
        path = path[:-1]
    if is_pd(value):
        res = pd_to_npy(value, path, append = append)
        res[_obj] = _pd_read_npy
    elif is_arr(value):
        fname = path + _npy 
        np_save(fname, value, append = append)
        return dict(_obj = _np_load, file = fname)        
    elif is_dict(value):
        return type(value)(**{k : npy_encode(v, '%s/%s'%(path,k), append = append) for k, v in value.items()})
    elif isinstance(value, (list, tuple)):
        return type(value)([npy_encode(v, '%s/%i'%(path,i), append = append) for i, v in enumerate(value)])
    else:
        return value

    


def csv_encode(value, path):
    """
    encodes a single DataFrame or a document containing dataframes into a an abject that can be decoded while saving dataframes into csv
    
    >>> path = 'c:/temp'
    >>> value = dict(key = 'a', data = dictable(a = [pd.Series([1,2,3]), pd.Series([4,5,6])], b = [1,2]), other = dict(df = pd.DataFrame(dict(a=[1,2,3], b= [4,5,6]))))
    >>> encoded = csv_encode(value, path)
    >>> assert encoded['data'].a[0]['path'] == 'c:/temp/data/a/0.csv'
    >>> assert encoded['other']['df']['path'] == 'c:/temp/other/df.csv'

    >>> decoded = decode(encoded)
    >>> assert eq(decoded, value)
    """
    if path.endswith(_csv):
        path = path[:-len(_csv)]
    if path.endswith('/'):
        path = path[:-1]
    if is_pd(value):
        return dict(_obj = _pd_read_csv, path = pd_to_csv(value, path))
    elif is_dict(value):
        return type(value)(**{k : csv_encode(v, '%s/%s'%(path,k)) for k, v in value.items()})
    elif isinstance(value, (list, tuple)):
        return type(value)([csv_encode(v, '%s/%i'%(path,i)) for i, v in enumerate(value)])
    else:
        return value
    
def npy_write(doc, root = None):
    """
    MongoDB is great for manipulating/searching dict keys/values. 
    However, the actual dataframes in each doc, we may want to save in a file system. 
    - The DataFrames are stored as bytes in MongoDB anyway, so they are not searchable
    - Storing in files allows other non-python/non-MongoDB users easier access, allowing data to be detached from app
    - MongoDB free version has limitations on size of document
    - file based system may be faster, especially if saved locally not over network
    - for data licensing issues, data must not sit on servers but stored on local computer

    Therefore, the doc encode will cycle through the elements in the doc. Each time it sees a pd.DataFrame/pd.Series, it will 
    - determine where to write it (with the help of the doc)
    - save it to a .parquet file

    >>> from pyg_base import *
    >>> from pyg_mongo import * 
    >>> db = mongo_table(db = 'temp', table = 'temp', pk = 'key', writer = 'c:/temp/%key.npy')         
    >>> a = pd.DataFrame(dict(a = [1,2,3], b= [4,5,6]), index = drange(2)); b = pd.DataFrame(np.random.normal(0,1,(3,2)), columns = ['a','b'], index = drange(2))
    >>> doc = dict(a = a, b = b, c = add_(a,b), key = 'b')
    >>> path ='c:/temp/%key'

    """
    if _root in doc:
        root  = doc[_root]
    if root is None and _db in doc and isinstance(doc[_db], partial) and _root in doc[_db].keywords:
        root = doc[_db].keywords[_root]
    if root is None:
        return doc
    path = root_path(doc, root)
    return npy_encode(doc, path)

            
def parquet_write(doc, root = None):
    """
    MongoDB is great for manipulating/searching dict keys/values. 
    However, the actual dataframes in each doc, we may want to save in a file system. 
    - The DataFrames are stored as bytes in MongoDB anyway, so they are not searchable
    - Storing in files allows other non-python/non-MongoDB users easier access, allowing data to be detached from app
    - MongoDB free version has limitations on size of document
    - file based system may be faster, especially if saved locally not over network
    - for data licensing issues, data must not sit on servers but stored on local computer

    Therefore, the doc encode will cycle through the elements in the doc. Each time it sees a pd.DataFrame/pd.Series, it will 
    - determine where to write it (with the help of the doc)
    - save it to a .parquet file

    >>> from pyg_base import *
    >>> from pyg_mongo import * 
    >>> db = mongo_table(db = 'temp', table = 'temp', pk = 'key', writer = 'c:/temp/%key.parquet')         
    >>> a = pd.DataFrame(dict(a = [1,2,3], b= [4,5,6]), index = drange(2)); b = pd.DataFrame(np.random.normal(0,1,(3,2)), columns = ['a','b'], index = drange(2))
    >>> doc = dict(a = a, b = b, c = add_(a,b), key = 'b')
    >>> path ='c:/temp/%key'

    """
    if _root in doc:
        root  = doc[_root]
    if root is None and _db in doc and isinstance(doc[_db], partial) and _root in doc[_db].keywords:
        root = doc[_db].keywords[_root]
    if root is None:
        return doc
    path = root_path(doc, root)
    return parquet_encode(doc, path)

def csv_write(doc, root = None):
    """
    MongoDB is great for manipulating/searching dict keys/values. 
    However, the actual dataframes in each doc, we may want to save in a file system. 
    - The DataFrames are stored as bytes in MongoDB anyway, so they are not searchable
    - Storing in files allows other non-python/non-MongoDB users easier access, allowing data to be detached from orignal application
    - MongoDB free version has limitations on size of document
    - file based system may be faster, especially if saved locally not over network
    - for data licensing issues, data must not sit on servers but stored on local computer

    Therefore, the doc encode will cycle through the elements in the doc. Each time it sees a pd.DataFrame/pd.Series, it will 
    - determine where to write it (with the help of the doc)
    - save it to a .csv file

    """
    if _root in doc:
        root  = doc[_root]
    if root is None and _db in doc and isinstance(doc[_db], partial) and _root in doc[_db].keywords:
        root = doc[_db].keywords[_root]
    if root is None:
        return doc
    path = root_path(doc, root)
    return csv_encode(doc, path)





from pyg_base import wrapper, getargs, as_list, getcallargs, cell_item
from pyg_mongo._table import mongo_table
from pyg_mongo._periodic_cell import periodic_cell
from functools import partial

class cell_cache(wrapper):
    """
    Rather than write this:

    >>> def function(a, b):
    >>>     return a + b        

    >>> db = partial(mongo_table, db = 'db', table = 'table', pk = ['key'])
    >>> periodic_cell(function, a =  1, b = 2, db = db, key = 'my key')
        
    You can write this:
        
    >>> f = cell_cache(function, db = 'db', table = 'table', pk = 'key')
    >>> f(a = 1, b = 2, key = 'my key')

    If we are interested just in values...
    
    >>> f = db_cache(function, db = 'db', table = 'table', pk = 'key')
    >>> assert f(a = 1, b = 2, key = 'my key') == 3

    :Parameters:
    ------------
    pk: str/list
        primary keys of the table, using the keyword arguments of the function. If missing, uses all keywords

    db: str
        name of database where data is to be stored

    table: str
        name of table/collection where data is stored. If not provided, defaults to the function's name

    cell: cell
        type of cell to use when caching the data

    cell_kwargs: dict
        parameters for the cell determining its operation, e.g. for periodic_cell, the periodicity

    :Example:
    ---------
    >>> from pyg import *
    >>> @db_cache
    >>> def f(a,b):
    >>>     return a+b
    
    >>> assert f(1,2) == 3

    >>> @cell_cache
    >>> def f(a,b):
    >>>     return a+b
    >>> assert f(1,2).data == 3

    >>> @cell_cache(pk = 'key')
    >>> def f(a,b):
    >>>     return a+b
    >>> f(1,2, key = 'key', go = 1)

    
    """
    def __init__(self, function = None, db = 'cache', table = None, pk = None, cell = periodic_cell, cell_kwargs = None, external = None):
        cell_kwargs  = cell_kwargs or {}
        super(cell_cache, self).__init__(function = function, pk = pk, db = db, table = table, cell = cell, cell_kwargs = cell_kwargs, external = external)

    @property
    def _pk(self):
        if self.pk is None and self.function is not None:
            args = getargs(self.function)
            self.pk = args
        return as_list(self.pk)

    @property
    def _external(self):
        if self.external is None and self.function is not None:
            args = getargs(self.function)
            self.external = [key for key in self._pk if key not in args]
        return self.external

    @property
    def _table(self):
        if self.table is None and self.function is not None:
            return self.function.__name__
        else:
            return self.table
    
    @property
    def _db(self):
        return partial(mongo_table, table = self._table, db = self.db, pk = self._pk)
    
    def wrapped(self, *args, **kwargs):
        go = kwargs.pop('go',0)
        mode = kwargs.pop('mode',0)
        external = self._external
        external_kwargs = {key : value for key, value in kwargs.items() if key in external}
        kwargs = {key : value for key, value in kwargs.items() if key not in external}
        callargs = getcallargs(self.function, *args, **kwargs)
        db = self._db
        external_kwargs.update(self.cell_kwargs)
        external_kwargs.update(callargs)
        return self.cell(self.function, db = db, **external_kwargs)(go = go, mode = mode)

class db_cache(cell_cache):
    def wrapped(self, *args, **kwargs):
        return cell_item(super(db_cache, self).wrapped(*args, **kwargs))
    





# -*- coding: utf-8 -*-

from pyg_mongo._q import Q, q, mdict
from pyg_mongo._types import is_collection, is_cursor
from pyg_mongo._base_reader import mongo_base_reader
from pyg_mongo._reader import mongo_reader
from pyg_mongo._cursor import mongo_cursor, mongo_pk_cursor

from pyg_mongo._table import mongo_table
from pyg_mongo._encoders import root_path, pd_to_csv, pd_read_csv, npy_encode, npy_write, parquet_encode, parquet_write, csv_encode, csv_write, encode

from pyg_mongo._db_cell import db_cell, db_load, db_save, get_cell, get_data, load_cell, load_data, cell_push, cell_pull
from pyg_mongo._periodic_cell import periodic_cell
from pyg_mongo._cache import db_cache, cell_cache



from pyg_base import as_list, is_strs, is_str, is_dict, is_int, dictable, sort, passthru
from pyg_mongo._q import _id, _doc, q, _set, _deleted
from pyg_mongo._base_reader import mongo_base_reader, _items1
import datetime

__all__ = ['mongo_reader']


class mongo_reader(mongo_base_reader):


    def _assert_one_or_none(self):
        n = self.count()
        if n>1:
            pk = self._pk
            if pk:
                for key in pk:
                    multiple_values = self.distinct(key)
                    if len(multiple_values) > 1:
                        raise KeyError('too many %s = %s found'%(key, multiple_values))
                self = self.dedup()
                n = self.count()
                if n <= 1:
                    return n
            raise ValueError('%s\nNon-unique %i documents %s... e.g. \n%s'%(self.collection, n, self._spec, self.read(slice(None,3,None))))
        else:
            return n

    def _assert_unique(self):
        n = self._assert_one_or_none()
        if n == 0:
            raise ValueError('%s\nNo documents %s'%(self.collection, self._spec))
        return self

    def find_one(self, doc = None, *args, **kwargs):
        res = self.find(*args, **kwargs)
        if doc:
            res = res.find(self._id(doc))
        return res._assert_unique()

    def read_one(self, doc = None, *args, **kwargs):
        reader = kwargs.pop('reader', None)
        return self.find_one(doc, *args, **kwargs).read(0, reader = reader)
                    

    def read(self, item = 0, reader = None):
        """
        reads the next document from the collection.

        :Parameters:
        ----------
        item : int, optional
            Please read the ith record. The default is 0.
        reader : callable/list of callables, optional
            When we read the document from the collection, we first transform them. 
            The default behaviour is to use pyg.base._encode.decode but you may pass reader = False to grab the raw data from mongo

        :Returns:
        -------
        document
            The document from Mongo
            
        item = 0

        """
        cursor = self.cursor
        if is_int(item):
            if item < 0:
                item = self.count() + item
            if item > 0:
                cursor = cursor.skip(item)
            doc = cursor.next()
            return self._read(doc, reader = reader)
        elif is_dict(item):
            res = self.find_one(item)
            return self._read(res.cursor.next(), reader = reader)
        elif isinstance(item, slice):
            return dictable([self._read(i, reader = reader) for i in self.cursor[self._item(item)]])
        elif isinstance(item, (list, range, tuple)):
            return [self.read(i, reader = reader) for i in item]

    
    def __getitem__(self, item):
        if is_str(item):
            return self.distinct(item)
        elif is_strs(as_list(item)):
            return self(projection = as_list(item))
        else:
            return self.read(item)
            
    
    def distinct(self, key):
        """
        returns the distinct cursor values of the key        
        """
        res = self.cursor.distinct(key)
        try:
            return sort(res)
        except TypeError:
            return res

    def docs(self, *keys, doc = _doc):
        """
        self[::] flattens the entire document.
        At times, we want to see the full documents, indexed by keys and docs does that.        
        returns a dictable with both keys and the document in the 'doc' column
        """
        keys = self._pk + as_list(keys)
        docs = [self._read(doc) for doc in self.cursor]
        res = dictable([{key: d.get(key) for key in keys} for d in docs])
        res[doc] = docs
        return res


    def create_index(self, *keys):
        keys = as_list(keys) or self._pk
        if len(keys) > 0:
            return self.collection.create_index(_items1(keys))
        return self

    def keys(self, item = 0):
        return self.read(item).keys()

    def count(self):
        return self.collection.count_documents(self._spec)

    __len__ = count    

    def __repr__(self):
        n = len(self)
        if n>0:
            return '%(t)s for %(c)s \n%(s)s %(p)s\ndocuments count: %(n)i \nfirst doc %(k)s'%dict(t = type(self), 
                                                                                          c = self.collection, 
                                                                                          s = self._spec, 
                                                                                          p = self._projection, 
                                                                                          n = n, 
                                                                                          k = self.keys())
        else:
            return '%(t)s for %(c)s \n%(s)s %(p)s\ndocuments count: 0'%dict(t = type(self), 
                                                                            c = self.collection, 
                                                                            s = self._spec, 
                                                                            p = self._projection)
            
                
        
    def _item(self, item):
        if isinstance(item, slice):
            return slice(self._item(item.start), self._item(item.stop), item.step)
        if  is_int(item) and item<0:
            return len(self) + item
        else:
            return item
        
    def __iter__(self):
        """
        When we iterate over documents we often change them, causing the cursor to change as well.
        This means often unexpected behaviour and in particular can lead to infinite loops too.
        We therefore choose to fix the _ids in advance.
        """
        _ids = self.collection.distinct(_id, self._spec)
        for i in _ids:
            yield self.inc({_id : i}).read(0)
            
    def __getattr__(self, key):
        if key.startswith('_'):
            return super(mongo_reader, self).__getattr__(key)
        else:            
            return self.distinct(key)
    

    def dedup(self):
        """
        Although in principle, if a single process reads/writes to Mongo, we should not get duplicates. 
        In practice, when multiple clients access the database, we occasionally get multiple records with the same primary keys.
        When this happens, we also end up with poor mongo _ids 

        Returns
        -------
        mongo_pk_cursor
            Hopefully, a table with unique keys.

        """
        if self.pk:
            pk = self._pk
            bad = self(projection = pk)[::].sort(_id).listby(pk).inc(lambda _id: len(_id)>1) 
            if len(bad):
                self.collection.update_many(q._id == sum(bad[lambda _id: _id[:-1]], []), {_set: {_deleted : datetime.datetime.now()}})
        return self
        




from pyg_base import logger, passthru, tree_update, dictable
from pyg_base import zipper, is_strs, is_dict, Dict, is_dictable, is_int, as_list, ulist
from pyg_mongo._q import _set, _id, _unset, _rename, _deleted, _data
from pyg_mongo._reader import mongo_reader
from pyg_mongo._base_reader import _pk, _items1, _dict1
import datetime


__all__ = ['mongo_cursor', 'mongo_pk_cursor']

class mongo_cursor(mongo_reader):
    """
    mongo_cursor is a souped-up combination of mongo.Cursor and mongo.Collection with a simple API.

    :Parameters:
    ----------------
    cursor : MongoDB cursor or MongoDB collection
        
    writer : True/False/string, optional
        The default is None.
        
        writer allows you to transform the data before saving it in Mongo. You can create a function yourself or use built-in options:
        
        - False: do nothing, save the document as is
        - True/None: use pyg.base.encode to encode objects. This will transform numpy array/dataframes into bytes that can be stored
        - '.csv': save dataframes into csv files and then save reference of these files to mongo
        - '.parquet' save dataframes into .parquet and np.ndarray into .npy files.
        
        For .csv and .parquet to work, you will need to specify WHERE the document is to be saved. This can be done either:
        
        - the document has a 'root' key, specifying the root.
        - you specify root by setting writer = 'c:/%name%surname.parquet'

    reader : callable or None, optional
        The default is None, using decode. Use reader = False to passthru
    query : dict, optional
        This is used to specify the Mongo query, e.g. q.a==1.
    **_ : 
    
        
    :Example:
    ---------
    >>> from pyg import *
    >>> cursor = mongo_table('test', 'test')
    >>> cursor.drop()

    ## insert some data
    
    >>> table = dictable(a = range(5)) * dictable(b = range(5))
    >>> cursor.insert_many(table)
    >>> cursor.set(c = lambda a, b: a * b)
    

    :filtering:
    -----------

    >>> assert len(cursor) == 25
    >>> assert len(cursor.find(a = 3)) == 5
    >>> assert len(cursor.exc(a = 3)) == 20
    >>> assert len(cursor.find(a = [3,2]).find(q.b<3)) == 6 ## can chain queries as well as use q to create complicated expressions
    
    :row access:
    -------------
    
    >>> cursor[0]
    
    {'_id': ObjectId('603aec85cd15e2c090c07b87'), 'a': 0, 'b': 0}
    
    >>> cursor[::] - '_id' == dictable(cursor) - '_id'
    
    >>> dictable[25 x 3]
    >>> a|b|c 
    >>> 0|0|0 
    >>> 0|1|0 
    >>> 0|2|0 
    >>> ...25 rows...
    >>> 4|2|8 
    >>> 4|3|12
    >>> 4|4|16
    
    :column access:
    ---------------
    >>> cursor[['a', 'b']]  ## just columns 'a' and 'b'
    >>> del cursor['c'] ## delete all c
    >>> cursor.set(c = lambda a, b: a * b)
    >>> assert cursor.find_one(a = 3, b = 2)[0].c == 6
    
    :Example: root specification
    ----------------------------
    >>> from pyg import *
    >>> t = mongo_table('test', 'test', writer = 'c:/temp/%name/%surname.parquet')
    >>> t.drop()
    >>> doc = dict(name = 'adam', surname = 'smith', ts = pd.Series(np.arange(10)))
    >>> t.insert_one(doc)
    >>> assert eq(pd_read_parquet('c:/temp/adam/smith/ts.parquet'), doc['ts'])
    >>> assert eq(t[0]['ts'], doc['ts'])
    >>> doc = dict(name = 'beth', surname = 'brown', a = np.arange(10))
    >>> t.insert_one(doc)
    
    Since mongo_cursor is too powerful, we also have a mongo_reader version which is read-only.
    
    
    """
    def delete_many(self, *args, **kwargs):
        """
        Equivalent to drop: deletes all documents the cursor currently points to.
        
        :Note: 
        ------
        If you want to drop a subset of the data, then use c.find(criteria).delete_many()

        :Returns:
        -------
        itself
        """
        target = self.inc(*args, **kwargs)
        n = target.count()
        spec = target._spec
        logger.info('INFO: deleting %i documents based on %s'%(n, spec))
        if n:
            target.collection.delete_many(spec)
        return self

        
    def delete_one(self, *args, **kwargs):
        """
        drops a specific record after verifying exactly one exists.

        :Parameters:
        ----------
        *args : query
        **kwargs : query

        :Returns:
        -------
        itself

        """
        c = self.find_one(*args, **kwargs)
        c.collection.delete_one(c._spec)
        return self

    drop = delete_many
    
    def _update_one(self, doc):
        """
        Uses the doc to find the existing document matching the doc. Then updates it. Throws an exception if no document found
        
        :Returns:
        ----------
        the updated document
        """
        update = self._write(doc)
        c = self.find_one(doc = update)
        update.pop(_id, None)
        self.collection.update_one(c._spec, {_set: update})
        return c[0]
    
    def update_one(self, doc, upsert = True):
        """
        - updates a document if an _id is present in doc.
        - insert a document if an _id is not present and upsert is true

        :Parameters:
        ----------
        doc : document
            doc to be upserted.
        upsert : bool, optional
            insert if no document present? The default is True.
        
        :Returns:
        -------
        doc
            document updated.

        """
        if upsert:
            return self.insert_one(doc)
        else:
            return self._update_one(doc)
    
    def update_many(self, doc, upsert  = False):
        """
        updates all documents in current cursor based on the doc. The two are equivalent:
        
        >>> cursot.update_many(doc)
        >>> collection.update_many(cursor._spec, { 'set' : update})

        :Parameters:
        ----------
        doc : dict of values to be updated

        :Returns:
        -------
        itself
        """
        update = self._write(doc)
        update.pop(_id, None)
        self.collection.update_many(self._spec, {_set : update})
        return self
    
    def __setitem__(self, key, value):
        update = dict(zipper(key, value))
        self.set(**update)
    
    def set(self, **kwargs):
        """
        updates all documents in current cursor based on the kwargs. 
        It is similar to update_many but supports also functions

        :Parameters:
        ----------
        kwargs: dict of values to be updated

        :Example:
        ---------
        >>> from pyg import *
        >>> t = mongo_table('test', 'test')
        >>> t = t.drop()
        >>> values = dictable(a = [1,2,3,4,], b = [5,6,7,8])
        >>> t = t.insert_many(values)
        >>> assert t[::]-'_id' == values

        >>> t.set(c = lambda a, b: a+b)
        >>> assert t[::]-'_id' == values(c = [6,8,10,12])
        >>> t.set(d = 1)
        >>> assert t[::]-'_id' == values(c =lambda a, b: a+b)(d = 1)

        :Returns:
        -------
        itself
        """
        static = {key: value for key, value in kwargs.items() if not callable(value)}
        if len(static) == len(kwargs):
            self.update_many(static)
        else:
            for row in self:
                tp = type(row)
                row = Dict(row) if not isinstance(row, Dict) else row
                doc = tp(row(**kwargs))
                self.update_one(doc)
        return self

    def rename(self, **kwargs):
        self.collection.update_many(self._spec, {_rename : kwargs})
        return self
    
    def __delitem__(self, item):
        if isinstance(item, int):
            self.collection.delete_one({_id : self[item][_id]})
        elif is_strs(item):
            self.collection.update_many(self._spec, {_unset: _dict1(item)})
        elif is_dict(item):
            self.find(item).delete_one()
    
    def delete(self, item):
        del self[item]
        return self
    
    def insert_one(self, doc):
        """
        inserts/updates a single document. 
        
        If the document ALREADY has _id in it, it updates that document
        If the document has no _id in it, it inserts it as a new document

        :Parameters:
        ----------
        doc : dict
            document.
            
        :Example:
        ---------
        >>> from pyg import *
        >>> t = mongo_table('test', 'test')
        >>> t = t.drop()
        >>> values = dictable(a = [1,2,3,4,], b = [5,6,7,8])
        >>> t = t.insert_many(values)
        
        :Example: used to update an existing document
        ---------------------------------------------
        >>> doc = t[0]
        >>> doc['c'] = 8
        >>> str(doc)
        >>> "{'_id': ObjectId('602d36150a5cd32717323197'), 'a': 1, 'b': 5, 'c': 8}"

        >>> t = t.insert_one(doc)
        >>> assert len(t) == 4        
        >>> assert t[0] == doc        

        :Example: used to insert 
        ------------------------
        >>> doc = Dict(a = 1, b = 8, c = 10)
        >>> t = t.insert_one(doc)
        >>> assert len(t) == 5
        >>> t.drop()
        """
        if _id in doc:
            return self._update_one(doc)
        else:
            res = doc.copy()
            new = self._write(doc)
            res[_id] = self.collection.insert_one(new).inserted_id
            return res

    def insert_many(self, table):
        """
        inserts multiple documents into the collection

        Parameters
        ----------
        table : sequence of documents
            list of dicts or dictable
        
        Returns
        -------
        mongo_cursor
        
        :Example: simple insertion
        ---------------------------
        >>> from pyg import *
        >>> t = mongo_table('test', 'test')
        >>> t = t.drop()
        >>> values = dictable(a = [1,2,3,4,], b = [5,6,7,8])
        >>> t = t.insert_many(values)
        >>> t[::]        

        >>> dictable[4 x 3]
        >>> _id                     |a|b
        >>> 602daee68c336f6429a77bdd|1|5
        >>> 602daee68c336f6429a77bde|2|6
        >>> 602daee68c336f6429a77bdf|3|7
        >>> 602daee68c336f6429a77be0|4|8

        :Example: update
        ----------------
        >>> table = t[::]
        >>> modified = table(b = lambda b: b**2)
        >>> t = t.insert_many(modified)

        Since each of the documents we uploaded already has an _id...

        >>> assert len(t) == 4
        >>> t[::]
        >>> dictable[4 x 3]
        >>> _id                     |a|b
        >>> 602daee68c336f6429a77bdd|1|25
        >>> 602daee68c336f6429a77bde|2|36
        >>> 602daee68c336f6429a77bdf|3|49
        >>> 602daee68c336f6429a77be0|4|64
    
        """
        with_ids = [self._write(doc) for doc in table if _id in doc]
        no_ids = [self._write(doc) for doc in table if _id not in doc]
        if len(no_ids)>0:
            self.collection.insert_many(no_ids)
        for doc in with_ids:
            spec = {_id: doc.pop(_id)}
            self.collection.update_one(spec, {_set: doc})
        return self

    def __add__(self, item):
        if is_dict(item) and not is_dictable(item):
            self.insert_one(item)
        else:
            self.insert_many(item)
        return self


    def __call__(self, **kwargs):
        callargs = self._callargs(**kwargs)
        obj = mongo_cursor if not callargs.get(_pk) else mongo_pk_cursor
        return obj(**callargs)
        

class mongo_pk_cursor(mongo_cursor):
    """
    The logic for reading with/without pk is similar
    for insertion though, the two are VERY different
    
    """
    def delete_one(self, doc = {}):
        return self.find_one(doc).delete_many()
    
    def delete_many(self):
        self.collection.update_many(self._spec, {_set: {_deleted : datetime.datetime.now()}})
        return self

    drop = delete_many

    @property
    def _pk(self):
        if not self.pk:
            raise ValueError('a mongo_pk_cursor must have some primary keys')
        res = ulist(sorted(set(as_list(self.pk))))
        return res
            

    def insert_one(self, doc):
        """
        marks the old document as deleted and inserts the new one.
        In actual fact, we maintain the old id when we insert the new document, The deleted document receives a new id 

        :Parameters:
        ----------------
        doc : document (dict)
            document to be inserted

        :Returns:
        -------
        _id 
            returned from Mongo

        """
        c = self.find(self._id(doc))
        n = c._assert_one_or_none()
        new = self._write(doc)
        new.pop(_id, None)
        if n == 1:
            old = c.read(0, reader = passthru)
            i = old.pop(_id)
            unset = _dict1(list(old))
            c.collection.update_one({_id : i}, {_unset: unset})
            c.collection.update_one({_id : i}, {_set: new})
            new[_id] = i
            old[_deleted] = datetime.datetime.now()
            self.collection.insert_one(old)
        else:
            missing_keys = [key for key in self._pk if key not in doc]
            if len(missing_keys)>0:
                raise ValueError('cannot save a new doc with primary keys %s missing'%missing_keys)
            new[_id] = self.collection.insert_one(new).inserted_id
        return new[_id]
    
    
    def _update_one(self, new):
        """
        receives a doc, returns updated doc
        """
        old = self.read(0, reader = passthru)
        if old is None:
            new[_id] = self.collection.insert_one(new).inserted_id
        else:
            i = old.pop(_id)
            new = tree_update(old, new)
            unset = _dict1(list(old))
            self.collection.update_one({_id : i}, {_unset: unset})
            self.collection.update_one({_id : i}, {_set: new})            
            new[_id] = i
            old[_deleted] = datetime.datetime.now()
            self.collection.insert_one(old)
        return new
    
    def update_one(self, doc, upsert = True):
        """
        updates an existing document

        :Parameters:
        ----------
        doc : document (dict or dict of dicts)
            DESCRIPTION.
        upsert : bool, optional
            If there isn't an existing document, should it insert doc as a new one?. The default is True.

        :Returns:
        ---------
        document
            new document post insertion/update.
            
        :Example:
        ---------
        >>> db = mongo_table('test', 'test', pk = ['name', 'surname'])
        >>> db.insert_one(dict(name = 'anna', surname = 'auburn', age = 39))        
        >>> db.update_one(dict(name = 'bobby', surname = 'benedict', age = 39))        

        >>> assert get_cell('test', 'test', surname = 'benedict')

        """
        new = self._write(doc)
        c = self.find(self._id(new))
        n = c._assert_one_or_none()
        if n == 1:
            return c._update_one(new)
        elif upsert:
            new = c._write(doc)
            new[_id] = c.collection.insert_one(self._write(new)).inserted_id
            return new ## always returns the encoded cell rather than the original
            
    def update_many(self, update, upsert = True):
        return type(update)([self.update_one(doc, upsert = upsert) for doc in update])

    def __setitem__(self, key, value):
        if isinstance(key, dict):
            doc = key.copy()
            doc.update(value if isinstance(value, dict) else {_data: value})
            self.update_one(doc, upsert = True)
        else:
            super(mongo_pk_cursor, self).__setitem__(key, value)
    


    def __delitem__(self, item):            
        if is_int(item):
            self.delete_one(self[item])
        elif is_strs(item):
            items = [i for i in as_list(item) if not i.startswith('_')]
            cannot_drop = self._pk & items
            if len(cannot_drop) > 0:
                raise ValueError('cannot drop primary keys %s'%cannot_drop)
            deleted = datetime.datetime.now()
            for doc in self:
                del doc[_id]
                doc.update({_deleted: deleted})
                self.collection.insert_one(self._write(doc))
            self.collection.update_many(self._spec, {_unset: _dict1(item)})
        elif isinstance(item, dict):
            self.delete_one(item)

    def delete(self, item):
        del self[item]
        return self

    def insert_many(self, table):
        for doc in table:
            self.insert_one(doc)
        return self
    
    def __add__(self, item):
        if isinstance(item, (list, dictable)):
            self.insert_many(item)
        elif is_dict(item):
            self.update_one(item, upsert = True)
        else:
            raise ValueError('can only add a dict, a dictable or a list of dicts')
        return self

    def set(self, **kwargs):
        rows = [row for row in self]
        for row in rows:
            doc = type(row)(Dict(row)(**kwargs))
            self.update_one(doc)
        return self
    
    @property
    def reset(self):
        return mongo_cursor(self.collection, writer = self.writer, reader = self.reader)

                    



from pyg_base import dt, dt_bump
from pyg_mongo._db_cell import db_cell, _updated

__all__ = ['periodic_cell']
_period = 'period'

class periodic_cell(db_cell):
    """
    periodic_cell inherits from db_cell its ability to save itself in MongoDb using its db members
    Its calculation schedule depends on when it was last updated. 
    
    :Example:
    ---------
    >>> from pyg import *
    >>> c = periodic_cell(lambda a: a + 1, a = 0)
    
    We now assert it needs to be calculated and calculate it...

    >>> assert c.run()
    >>> c = c.go()
    >>> assert c.data == 1
    >>> assert not c.run()
    
    Now let us cheat and tell it, it was last run 3 days ago...
    
    >>> c.updated = dt(-3)
    >>> assert c.run()
            
    """
    def __init__(self, function = None, output = None, db = None, period = '1b', updated = None, **kwargs):
        self[_updated] = updated
        self[_period] = period 
        super(periodic_cell, self).__init__(function, output = output, db = db, **kwargs)
            
    def run(self):
        time = dt()
        if self[_updated] is None or dt_bump(self[_updated], self[_period]) < time:
            return True
        return super(periodic_cell, self).run() 

